{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "gHoQThpObPSq"
      },
      "source": [
        "# 第六次作业\n",
        "\n",
        "在本次作业中，大家需要完成GRAND模型的训练，重点体会一致性正则化（Consistency Regularization）的用法，并在cora数据集上进行测试。\n",
        "\n",
        "本作业需要安装[CogDL](https://github.com/THUDM/cogdl)：pip install cogdl\n",
        "\n",
        "如需使用gpu版，请先安装gpu版本的[PyTorch](https://pytorch.org/get-started/locally/)，再安装cogdl。\n",
        "\n",
        "本作业由智谱GNN中心及课程团队筹备，由CogDL团队提供技术支持。\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### 1. GRAND模型\n",
        "阅读GRAND模型的实现。"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {},
      "outputs": [],
      "source": [
        "import torch\n",
        "import torch.nn as nn\n",
        "import torch.nn.functional as F\n",
        "\n",
        "from cogdl.utils import spmm\n",
        "\n",
        "\n",
        "class Grand(nn.Module):\n",
        "    \"\"\"\n",
        "    Implementation of GRAND in paper `\"Graph Random Neural Networks for Semi-Supervised Learning on Graphs\"`\n",
        "    <https://arxiv.org/abs/2005.11079>\n",
        "    Parameters\n",
        "    ----------\n",
        "    nfeat : int\n",
        "        Size of each input features.\n",
        "    nhid : int\n",
        "        Size of hidden features.\n",
        "    nclass : int\n",
        "        Number of output classes.\n",
        "    input_droprate : float\n",
        "        Dropout rate of input features.\n",
        "    hidden_droprate : float\n",
        "        Dropout rate of hidden features.\n",
        "    use_bn : bool\n",
        "        Using batch normalization.\n",
        "    dropnode_rate : float\n",
        "        Rate of dropping elements of input features\n",
        "    order : int\n",
        "        Order of adjacency matrix\n",
        "    alpha : float\n",
        "    \"\"\"\n",
        "\n",
        "    def __init__(\n",
        "        self, nfeat, nhid, nclass, input_droprate, hidden_droprate, use_bn, dropnode_rate, order, alpha,\n",
        "    ):\n",
        "        super(Grand, self).__init__()\n",
        "        self.layer1 = nn.Linear(nfeat, nhid)\n",
        "        self.layer2 = nn.Linear(nhid, nclass)\n",
        "        self.input_droprate = input_droprate\n",
        "        self.hidden_droprate = hidden_droprate\n",
        "        self.bn1 = nn.BatchNorm1d(nfeat)\n",
        "        self.bn2 = nn.BatchNorm1d(nhid)\n",
        "        self.use_bn = use_bn\n",
        "        self.order = order\n",
        "        self.dropnode_rate = dropnode_rate\n",
        "        self.alpha = alpha\n",
        "\n",
        "    def drop_node(self, x):\n",
        "        n = x.shape[0]\n",
        "        drop_rates = torch.ones(n) * self.dropnode_rate\n",
        "        if self.training:\n",
        "            masks = torch.bernoulli(1.0 - drop_rates).unsqueeze(1)\n",
        "            x = masks.to(x.device) * x\n",
        "        else:\n",
        "            x = x * (1.0 - self.dropnode_rate)\n",
        "        return x\n",
        "\n",
        "    def rand_prop(self, graph, x):\n",
        "        x = self.drop_node(x)\n",
        "        y = x\n",
        "        for i in range(self.order):\n",
        "            x = spmm(graph, x).detach_()\n",
        "            y.add_(x)\n",
        "        return y.div_(self.order + 1.0).detach_()\n",
        "\n",
        "    def normalize_x(self, x):\n",
        "        row_sum = x.sum(1)\n",
        "        row_inv = row_sum.pow_(-1)\n",
        "        row_inv.masked_fill_(row_inv == float(\"inf\"), 0)\n",
        "        x = x * row_inv[:, None]\n",
        "        return x\n",
        "\n",
        "    def forward(self, graph):\n",
        "        graph.sym_norm()\n",
        "        x = graph.x\n",
        "        x = self.normalize_x(x)\n",
        "        x = self.rand_prop(graph, x)\n",
        "        if self.use_bn:\n",
        "            x = self.bn1(x)\n",
        "        x = F.dropout(x, self.input_droprate, training=self.training)\n",
        "        x = F.relu(self.layer1(x))\n",
        "        if self.use_bn:\n",
        "            x = self.bn2(x)\n",
        "        x = F.dropout(x, self.hidden_droprate, training=self.training)\n",
        "        x = self.layer2(x)\n",
        "        return x"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "OAPbTveYjONP"
      },
      "source": [
        "### 2. 从cogdl中加载cora数据集（x表示特征，y表示标签，mask表示训练/验证/测试集的划分）"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 2,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "61aEVO2ySq27",
        "outputId": "6da332b4-d5aa-4baf-d95e-4dd5b4b291da"
      },
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Graph(x=[2708, 1433], y=[2708], train_mask=[2708], val_mask=[2708], test_mask=[2708], edge_index=[2, 13264], edge_weight=[13264])\n"
          ]
        }
      ],
      "source": [
        "from cogdl.datasets import build_dataset_from_name\n",
        "\n",
        "dataset = build_dataset_from_name(\"cora\")\n",
        "data = dataset[0]\n",
        "data.add_remaining_self_loops()\n",
        "print(data)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "FcRyHUuIjhfT"
      },
      "source": [
        "### 3. 使用GRAND模型进行训练"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "#### 3.1 GRAND的训练逻辑"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 3,
      "metadata": {},
      "outputs": [],
      "source": [
        "import torch\n",
        "import torch.nn as nn\n",
        "import torch.nn.functional as F\n",
        "\n",
        "\n",
        "class GrandModelWrapper(object):\n",
        "    \"\"\"\n",
        "    sample : int\n",
        "        Number of augmentations for consistency loss\n",
        "    temperature : float\n",
        "        Temperature to sharpen predictions.\n",
        "    lmbda : float\n",
        "         Proportion of consistency loss of unlabelled data\n",
        "    \"\"\"\n",
        "\n",
        "    def __init__(self, model, sample=2, temperature=0.5, lmbda=0.5):\n",
        "        super(GrandModelWrapper, self).__init__()\n",
        "        self.model = model\n",
        "        self.sample = sample\n",
        "        self.temperature = temperature\n",
        "        self.lmbda = lmbda\n",
        "\n",
        "    def train_step(self, graph):\n",
        "        output_list = []\n",
        "        for i in range(self.sample):\n",
        "            output_list.append(self.model(graph))\n",
        "        loss_train = 0.0\n",
        "        for output in output_list:\n",
        "            loss_train += F.cross_entropy(output[graph.train_mask], graph.y[graph.train_mask])\n",
        "        loss_train = loss_train / self.sample\n",
        "\n",
        "        output_list = [F.log_softmax(x, dim=-1) for x in output_list]\n",
        "        loss_consis = self.consistency_loss(output_list, graph.train_mask)\n",
        "\n",
        "        return loss_train + loss_consis\n",
        "\n",
        "    def consistency_loss(self, logps, train_mask):\n",
        "        temp = self.temperature\n",
        "        ps = [torch.exp(p)[~train_mask] for p in logps]\n",
        "\n",
        "        ###################\n",
        "        ##### 作业填空 #####\n",
        "        ###################\n",
        "        # 根据模型的多组输出计算GRAND论文中的一致性正则化损失函数（参考GRAND论文的公式2/3）\n",
        "        # 整体流程包括：1）Average；2）Sharpening；3）L2 distance\n",
        "        # 注：ps中保存了模型对于非训练集中的结果的多组预测结果\n",
        "\n",
        "        loss = ...\n",
        "\n",
        "        return self.lmbda * loss"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "#### 3.2 训练GRAND模型"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 4,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "pr5lZC2eWCYm",
        "outputId": "c42b59ef-9637-4bac-e1f1-79a3f9f6228a"
      },
      "outputs": [
        {
          "name": "stderr",
          "output_type": "stream",
          "text": [
            "  0%|          | 0/100 [00:00<?, ?it/s]"
          ]
        },
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Model Parameters: 49049\n"
          ]
        },
        {
          "name": "stderr",
          "output_type": "stream",
          "text": [
            "Epoch: 099, Train Loss: 0.8854, Val Loss: 0.8261, Val Acc: 0.8100: 100%|██████████| 100/100 [07:21<00:00,  4.42s/it]\n"
          ]
        },
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Val Acc 0.81\n",
            "Test Acc 0.825\n"
          ]
        }
      ],
      "source": [
        "import copy\n",
        "from tqdm import tqdm\n",
        "\n",
        "\n",
        "def accuracy(y_pred, y_true):\n",
        "    y_true = y_true.squeeze().long()\n",
        "    preds = y_pred.max(1)[1].type_as(y_true)\n",
        "    correct = preds.eq(y_true).double()\n",
        "    correct = correct.sum().item()\n",
        "    return correct / len(y_true)\n",
        "\n",
        "def train(hidden_size=32):\n",
        "    hidden_size = hidden_size\n",
        "    model = Grand(\n",
        "        data.x.shape[1], hidden_size, data.y.max() + 1, \n",
        "        input_droprate=0.5, hidden_droprate=0.5, use_bn=False,\n",
        "        dropnode_rate=0.5, order=5, alpha=0.2,\n",
        "    )\n",
        "    print(\"Model Parameters:\", sum(p.numel() for p in model.parameters()))\n",
        "\n",
        "    if torch.cuda.is_available():\n",
        "        device = torch.device(\"cuda\")\n",
        "        model = model.to(device)\n",
        "        data.apply(lambda x: x.to(device))\n",
        "\n",
        "    mw = GrandModelWrapper(model, sample=4, temperature=0.5, lmbda=1.0)\n",
        "\n",
        "    optimizer = torch.optim.AdamW(model.parameters(), lr=0.01)\n",
        "    epoch_iter = tqdm(range(100), position=0, leave=True)\n",
        "    best_model = None\n",
        "    best_acc = 0\n",
        "    for epoch in epoch_iter:\n",
        "        model.train()\n",
        "        optimizer.zero_grad()\n",
        "\n",
        "        loss = mw.train_step(data)\n",
        "        loss.backward()\n",
        "\n",
        "        optimizer.step()\n",
        "        train_loss = loss.item()\n",
        "\n",
        "        model.eval()\n",
        "        with torch.no_grad():\n",
        "            logits = model(data)\n",
        "            val_loss = F.cross_entropy(logits[data.val_mask], data.y[data.val_mask]).item()\n",
        "            val_acc = accuracy(logits[data.val_mask], data.y[data.val_mask])\n",
        "            if val_acc > best_acc:\n",
        "                best_acc = val_acc\n",
        "                best_model = copy.deepcopy(model)\n",
        "\n",
        "        epoch_iter.set_description(f\"Epoch: {epoch:03d}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}, Val Acc: {val_acc:.4f}\")\n",
        "\n",
        "    with torch.no_grad():\n",
        "        logits = best_model(data)\n",
        "        val_acc = accuracy(logits[data.val_mask], data.y[data.val_mask])\n",
        "        test_acc = accuracy(logits[data.test_mask], data.y[data.test_mask])\n",
        "    print(\"Val Acc\", val_acc)\n",
        "    print(\"Test Acc\", test_acc)\n",
        "\n",
        "train()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### 4. 开放讨论\n",
        "谈谈你对一致性正则化(Consistency Regularization)的理解。你觉得在什么情况下一致性正则化能够带来效果提升？（可以将你的思考发到讨论区中：https://discuss.cogdl.ai/t/topic/83 ）\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": []
    }
  ],
  "metadata": {
    "accelerator": "GPU",
    "colab": {
      "collapsed_sections": [],
      "name": "write-your-first-gcn",
      "provenance": []
    },
    "kernelspec": {
      "display_name": "Python 3.6.15 ('py3.6')",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.15"
    },
    "vscode": {
      "interpreter": {
        "hash": "7703d8cc28a94efd96474faf641471b2ac55415e3c97cc8df6cd7b51cf78a952"
      }
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}
